@@ -11,15 +11,19 @@ of the :class:`EvaluableGraph` class. The external library `graph-scheduler
 conditional constraints.
 
 """
+import copy
 import functools
 import inspect
+import itertools
 import os
 import re
 import sys
+import math
+
+import attr
 import sympy
 import numpy as np
 
-
 import graph_scheduler
 import onnxruntime
 
@@ -31,7 +35,7 @@ from modelspec.utils import _params_info, _val_info
 from modelspec.utils import FORMAT_NUMPY
 
 from collections import OrderedDict
-from typing import Union, List, Dict, Optional, Any
+from typing import Union, List, Dict, Optional, Any, Tuple
 from modeci_mdf.mdf import (
     Function,
     Graph,
@@ -52,6 +56,9 @@ FORMAT_DEFAULT = FORMAT_NUMPY
 KNOWN_PARAMETERS = ["constant"]
 
 
+time_scale_str_regex = r"(TimeScale)?\.(.*)"
+
+
 def evaluate_expr(
     expr: Union[str, List[str], np.ndarray, "tf.tensor"] = None,
     func_params: Dict[str, Any] = None,
@@ -78,7 +85,7 @@ def evaluate_expr(
     if type(e) == str and e not in KNOWN_PARAMETERS:
         raise Exception(
             "Error! Could not evaluate expression [%s] with params %s, returned [%s] which is a %s"
-            % (expr, _params_info(func_params), e, type(e))
+            % (expr, _params_info(func_params, multiline=True), e, type(e))
         )
     return e
 
@@ -303,15 +310,11 @@ class EvaluableFunction:
         # func_val  = self.function.value
 
         if self.function.function:
-
             for f in mdf_functions:
-
-                if f in self.function.function.keys():
-
+                if f == self.function.function:
                     expr = create_python_expression(
                         mdf_functions[f]["expression_string"]
                     )
-
                     break
 
         if expr is None:
@@ -328,27 +331,21 @@ class EvaluableFunction:
                 "    Evaluating %s with %s, i.e. [%s]"
                 % (self.function, _params_info(func_params), expr)
             )
-        if self.function.function:
 
-            for f in mdf_functions:
-
-                if f in self.function.function.keys():
-
-                    for arg in self.function.function[f].keys():
-
-                        func_params[arg] = evaluate_expr(
-                            self.function.function[f][arg],
-                            func_params,
-                            verbose=False,
-                            array_format=array_format,
+        if self.function.args is not None:
+            for arg in self.function.args:
+                func_params[arg] = evaluate_expr(
+                    self.function.args[arg],
+                    func_params,
+                    verbose=False,
+                    array_format=array_format,
+                )
+                if self.verbose:
+                    print(
+                        "      Arg: {} became: {}".format(
+                            arg, _val_info(func_params[arg])
                         )
-                        if self.verbose:
-                            print(
-                                "      Arg: {} became: {}".format(
-                                    arg, _val_info(func_params[arg])
-                                )
-                            )
-                    break
+                    )
 
         # If this is an ONNX operation, evaluate it without modelspec.
 
@@ -395,6 +392,7 @@ class EvaluableParameter:
         self.verbose = verbose
         self.parameter = parameter
 
+        """
         if self.parameter.default_initial_value is not None:
             if is_number(self.parameter.default_initial_value):
 
@@ -410,7 +408,8 @@ class EvaluableParameter:
                 except Exception:
                     self.curr_value = self.parameter.default_initial_value
         else:
-            self.curr_value = None
+            self.curr_value = None"""
+        self.curr_value = None
 
     def get_current_value(
         self, parameters: Dict[str, Any], array_format: str = FORMAT_DEFAULT
@@ -419,22 +418,30 @@ class EvaluableParameter:
         Get the current value of the parameter; evaluates the expression if needed.
 
         Args:
-            parameters: a dictionary  of parameters and their values that may or may not be needed to evaluate this
+            parameters: a dictionary of parameters and their values that may or may not be needed to evaluate this
                 parameter.
             array_format: The array format to use (either :code:`'numpy'` or :code:`tensorflow'`).
 
         Returns:
             The evaluated value of the parameter.
         """
-
         # FIXME: Shouldn't this just call self.evaluate, seems like there is redundant code here?
         if self.curr_value is None:
-
-            if self.parameter.value is not None:
+            if (
+                self.parameter.value is not None
+                or self.parameter.default_initial_value is not None
+            ):
                 if self.parameter.is_stateful():
+                    if self.verbose:
+                        print(f"    Initial eval of <{self.parameter.summary()}>  ")
 
                     if self.parameter.default_initial_value is not None:
-                        return self.parameter.default_initial_value
+                        return evaluate_expr(
+                            self.parameter.default_initial_value,
+                            parameters,
+                            verbose=False,
+                            array_format=array_format,
+                        )
                     else:
                         return self.DEFAULT_INIT_VALUE
                 else:
@@ -466,7 +473,7 @@ class EvaluableParameter:
         Evaluate the parameter and store the result in the :code:`curr_value` attribute.
 
         Args:
-            parameters: a dictionary  of parameters and their values that may or may not be needed to evaluate this
+            parameters: a dictionary of parameters and their values that may or may not be needed to evaluate this
                 parameter.
             time_increment: a floating point value specifying the timestep size, only used for :code:`time_derivative`
                 parameters
@@ -478,98 +485,133 @@ class EvaluableParameter:
         if self.verbose:
             print(
                 "    Evaluating {} with {} ".format(
-                    self.parameter, _params_info(parameters)
+                    self.parameter.summary(), _params_info(parameters)
                 )
             )
 
-        if self.parameter.value is not None:
-
-            self.curr_value = evaluate_expr(
-                self.parameter.value,
-                parameters,
-                verbose=False,
-                array_format=array_format,
-            )
-        elif self.parameter.function:
-            expr = None
-            for f in mdf_functions:
-                if self.parameter.function == f:
-                    expr = create_python_expression(
-                        mdf_functions[f]["expression_string"]
-                    )
-            if not expr:
-
-                expr = self.parameter.function
-                # raise "Unknown function: {}. Known functions: {}".format(
-                #    self.parameter.function,
-                #    mdf_functions.keys,
-                # )
-
-            func_params = {}
-            func_params.update(parameters)
-            if self.verbose:
-                print(
-                    "    Evaluating %s with %s, i.e. [%s]"
-                    % (self.parameter, _params_info(func_params), expr)
-                )
-            for arg in self.parameter.args:
-                func_params[arg] = evaluate_expr(
-                    self.parameter.args[arg],
-                    func_params,
+        continue_eval = True
+        if len(self.parameter.conditions) > 0:
+            for condition in self.parameter.conditions:
+                test = evaluate_expr(
+                    condition.test,
+                    parameters,
                     verbose=False,
                     array_format=array_format,
                 )
+                if test == True:
+                    self.curr_value = evaluate_expr(
+                        condition.value,
+                        parameters,
+                        verbose=False,
+                        array_format=array_format,
+                    )
+                    continue_eval = False
+                    break
+
                 if self.verbose:
                     print(
-                        "      Arg: {} became: {}".format(
-                            arg, _val_info(func_params[arg])
-                        )
+                        " --- %s: %s = %s (continuing: %s)"
+                        % (condition.id, condition.test, test, continue_eval)
                     )
 
-            # If this is an ONNX operation, evaluate it without modelspec.
-            if "onnx_ops." in expr:
+        if continue_eval:
+            if self.parameter.value is not None:
+
+                self.curr_value = evaluate_expr(
+                    self.parameter.value,
+                    parameters,
+                    verbose=False,
+                    array_format=array_format,
+                )
+            elif self.parameter.function:
+                expr = None
+                for f in mdf_functions:
+                    if self.parameter.function == f:
+                        expr = create_python_expression(
+                            mdf_functions[f]["expression_string"]
+                        )
+                if not expr:
+
+                    expr = self.parameter.function
+                    # raise "Unknown function: {}. Known functions: {}".format(
+                    #    self.parameter.function,
+                    #    mdf_functions.keys,
+                    # )
+
+                func_params = {}
+                func_params.update(parameters)
                 if self.verbose:
-                    print(f"{self.parameter.id} is evaluating ONNX function {expr}")
-                self.curr_value = evaluate_onnx_expr(
-                    expr,
-                    # parameters get overridden by self.parameter.args
-                    {**parameters, **self.parameter.args},
-                    func_params,
-                    self.verbose,
-                )
-            elif "actr." in expr:
-                actr_function = getattr(actr_funcs, expr.split("(")[0].split(".")[-1])
-                self.curr_value = actr_function(
-                    *[func_params[arg] for arg in self.parameter.args]
-                )
+                    print(
+                        "    Evaluating %s with %s, i.e. [%s]"
+                        % (self.parameter, _params_info(func_params), expr)
+                    )
+                for arg in self.parameter.args:
+                    func_params[arg] = evaluate_expr(
+                        self.parameter.args[arg],
+                        func_params,
+                        verbose=False,
+                        array_format=array_format,
+                    )
+                    if self.verbose:
+                        print(
+                            "      Arg: {} became: {}".format(
+                                arg, _val_info(func_params[arg])
+                            )
+                        )
+
+                # If this is an ONNX operation, evaluate it without modelspec.
+                if "onnx_ops." in expr:
+                    if self.verbose:
+                        print(f"{self.parameter.id} is evaluating ONNX function {expr}")
+                    self.curr_value = evaluate_onnx_expr(
+                        expr,
+                        # parameters get overridden by self.parameter.args
+                        {**parameters, **self.parameter.args},
+                        func_params,
+                        self.verbose,
+                    )
+                elif "actr." in expr:
+                    actr_function = getattr(
+                        actr_funcs, expr.split("(")[0].split(".")[-1]
+                    )
+                    self.curr_value = actr_function(
+                        *[func_params[arg] for arg in self.parameter.args]
+                    )
+                else:
+                    self.curr_value = evaluate_expr(
+                        expr,
+                        func_params,
+                        verbose=self.verbose,
+                        array_format=array_format,
+                    )
             else:
-                self.curr_value = evaluate_expr(
-                    expr, func_params, verbose=self.verbose, array_format=array_format
-                )
-        else:
-            if time_increment == None:
+                if time_increment == None:
 
-                self.curr_value = evaluate_expr(
-                    self.parameter.default_initial_value,
-                    parameters,
-                    verbose=False,
-                    array_format=array_format,
-                )
+                    self.curr_value = evaluate_expr(
+                        self.parameter.default_initial_value,
+                        parameters,
+                        verbose=False,
+                        array_format=array_format,
+                    )
 
-            else:
-                td = evaluate_expr(
-                    self.parameter.time_derivative,
-                    parameters,
-                    verbose=False,
-                    array_format=array_format,
-                )
+                else:
+                    td = evaluate_expr(
+                        self.parameter.time_derivative,
+                        parameters,
+                        verbose=False,
+                        array_format=array_format,
+                    )
 
-                self.curr_value += td * time_increment
+                    self.curr_value += td * time_increment
 
         if self.verbose:
             print(
                 "    Evaluated %s with %s \n       =\t%s"
-                % (self.parameter, _params_info(parameters), _val_info(self.curr_value))
+                % (
+                    self.parameter.summary(),
+                    _params_info(parameters),
+                    _val_info(self.curr_value),
+                )
             )
 
         return self.curr_value
@@ -586,6 +628,7 @@ class EvaluableOutput:
     def __init__(self, output_port: OutputPort, verbose: Optional[bool] = False):
         self.verbose = verbose
         self.output_port = output_port
+        self.curr_value = None
 
     def evaluate(
         self,
@@ -662,7 +705,7 @@ class EvaluableInput:
         """
         if self.verbose:
             print(
-                "    Evaluated %s with %s =\t%s"
+                "    Evaluated %s with params %s =\t%s"
                 % (
                     self.input_port,
                     _params_info(parameters),
@@ -766,7 +809,9 @@ class EvaluableNode:
                     )
                 else:
                     all_funcs.append(f)
+
         all_params_to_check = [p for p in node.parameters]
+
         if self.verbose:
             print("all_params_to_check: %s" % all_params_to_check)
 
@@ -842,14 +887,19 @@ class EvaluableNode:
             rop = EvaluableOutput(op, self.verbose)
             self.evaluable_outputs[op.id] = rop
 
-    def initialize(self):
-        pass
-
     def evaluate(
         self,
         time_increment: Union[int, float] = None,
         array_format: str = FORMAT_DEFAULT,
     ):
+        """
+        Evaluate the Node for one time-step
+
+        Args:
+            time_increment: The time-increment to use for this evaluation.
+            array_format: The format to use for arrays.
+
+        """
 
         if self.verbose:
             print(
@@ -885,19 +935,29 @@ class EvaluableNode:
         for eop in self.evaluable_outputs:
             self.evaluable_outputs[eop].evaluate(curr_params, array_format=array_format)
 
-    def get_output(self, id: str) -> Union[int, np.ndarray]:
+    def get_output(self, id: str = None) -> Union[int, np.ndarray, Tuple]:
         """Get value at output port for given output port's id
 
         Args:
-            id: Unique identifier of the output port
+            id: Unique identifier of the output port. If None, return a tuple for all output ports.
 
         Returns:
-            value at the output port
+            value at the output port. If id is None, return all outputs as a tuple. If there is only
+            one output, return just its value.
 
         """
-        for rop in self.evaluable_outputs:
-            if rop == id:
-                return self.evaluable_outputs[rop].curr_value
+        if id is not None:
+            for rop in self.evaluable_outputs:
+                if rop == id:
+                    return self.evaluable_outputs[rop].curr_value
+        else:
+            outputs = tuple(
+                self.evaluable_outputs[rop].curr_value for rop in self.evaluable_outputs
+            )
+            if len(outputs) == 1:
+                return outputs[0]
+            else:
+                return outputs
 
 
 class EvaluableGraph:
@@ -916,7 +976,10 @@ class EvaluableGraph:
         self.graph = graph
         self.enodes = {}
         self.root_nodes = []
+        self.output_nodes = []
 
+        # Get the root (input nodes) of the graph. We will assume all are root nodes
+        # and then remove those that have edges to them.
         for node in graph.nodes:
             if self.verbose:
                 print("\n  Init node: %s" % node.id)
@@ -946,25 +1009,48 @@ class EvaluableGraph:
                 evaluated_nodes.append(edge.receiver)
 
         if self.graph.conditions is not None:
-            conditions = {
-                self.graph.get_node(node): self.parse_condition(cond)
-                for node, cond in self.graph.conditions.node_specific.items()
-            }
+            if self.graph.conditions.node_specific is None:
+                conditions = {}
+            else:
+                conditions = {
+                    self.graph.get_node(node): self.parse_condition(cond)
+                    for node, cond in self.graph.conditions.node_specific.items()
+                }
 
-            termination_conds = {
-                scale: self.parse_condition(cond)
-                for scale, cond in self.graph.conditions.termination.items()
-            }
+            termination_conds = {}
+            if self.graph.conditions.termination is not None:
+                for scale, cond in self.graph.conditions.termination.items():
+                    cond = self.parse_condition(cond)
+
+                    # check for TimeScale in form of enum or equivalent unambiguous strings
+                    try:
+                        scale = re.match(time_scale_str_regex, scale).groups()[1]
+                    except (AttributeError, IndexError, TypeError):
+                        pass
+
+                    try:
+                        termination_conds[graph_scheduler.TimeScale[scale]] = cond
+                    except KeyError:
+                        termination_conds[scale] = cond
         else:
             conditions = {}
             termination_conds = {}
-
         self.scheduler = graph_scheduler.Scheduler(
             graph=self.graph.dependency_dict,
             conditions=conditions,
             termination_conds=termination_conds,
         )
 
+        # We also need to get the output nodes
+        self.output_nodes = [
+            n
+            for n in self.graph.nodes
+            if n.id not in (e.sender for e in self.graph.edges)
+        ]
+
+        # Lets also get the corresponding EvaluableNode for outputs
+        self.output_enodes = [self.enodes[n.id] for n in self.output_nodes]
+
     def evaluate(
         self,
         time_increment: Union[int, float] = None,
@@ -1027,7 +1113,9 @@ class EvaluableGraph:
                 )
 
         if self.verbose:
-            print("Trial terminated")
+            print("Order of execution of nodes\n")
+            print(list(self.scheduler.run()))
+            print("\n Trial terminated")
 
     def evaluate_edge(
         self,
@@ -1062,11 +1150,14 @@ class EvaluableGraph:
                     _val_info(weight),
                 )
             )
-        input_value = value if weight == 1 else value * weight
+        if (type(weight) == int or type(weight) == float) and weight == 1:
+            input_value = value
+        else:
+            input_value = weight * value
         post_node.evaluable_inputs[edge.receiver_port].set_input_value(input_value)
 
     def parse_condition(
-        self, condition: Union[Condition, dict]
+        self, condition: Union[Condition, Dict]
     ) -> graph_scheduler.Condition:
         """Convert the condition in a specific format
 
@@ -1078,7 +1169,17 @@ class EvaluableGraph:
 
         """
 
-        def substitute_condition_arguments(condition, condition_type):
+        def get_custom_parameter_getter(eobj):
+            def getter(dependency, parameter):
+                res = eobj.curr_value
+                if res is None:
+                    return 0
+                else:
+                    return res
+
+            return getter
+
+        def update_condition_arguments(args, condition_type):
             # mdf format prefers the key name for all conditions that use it or
             # its value as an __init__ argument
             combined_condition_arguments = {"dependencies": "dependency"}
@@ -1086,16 +1187,58 @@ class EvaluableGraph:
 
             for preferred, actual in combined_condition_arguments.items():
                 if (
-                    preferred in condition.args
+                    preferred in condition.kwargs
                     and preferred not in sig.parameters
                     and actual in sig.parameters
                 ):
-                    condition.args[actual] = condition.args[preferred]
-                    del condition.args[preferred]
+                    args[actual] = args[preferred]
+                    del args[preferred]
+
+            if "custom_parameter_getter" in sig.parameters:
+                try:
+                    dependency = args["dependency"]
+                    parameter = args["parameter"]
+                except KeyError as e:
+                    raise ValueError(
+                        "Threshold condition did not specify dependency or parameter"
+                    ) from e
+
+                # self is EvaluableGraph condition is running under,
+                enode = self.enodes[dependency.id]
+                evaluable_objects = [
+                    enode.evaluable_inputs,
+                    enode.evaluable_parameters,
+                    enode.evaluable_functions,
+                    enode.evaluable_outputs,
+                ]
+                valid_parameters = itertools.chain.from_iterable(
+                    [obj for obj in evaluable_objects]
+                )
+                args["custom_parameter_validator"] = (
+                    lambda dependency, parameter: parameter in valid_parameters
+                )
+                # assumes a unique parameter id among evaluable objects for
+                # dependency, or will resolve in favor of first with that id
+                for obj in evaluable_objects:
+                    try:
+                        obj = obj[parameter]
+                    except KeyError:
+                        pass
+                    else:
+                        args["custom_parameter_getter"] = get_custom_parameter_getter(
+                            obj
+                        )
+                        break
+                else:
+                    raise ValueError(
+                        f"No {parameter} evaluable object for {dependency}, options: {list(valid_parameters)}"
+                    )
+
+            return args
 
         # if specified as dict
         try:
-            args = condition["args"]
+            args = condition["kwargs"]
         except (TypeError, KeyError):
             args = {}
 
@@ -1105,7 +1248,7 @@ class EvaluableGraph:
             pass
 
         cond_type = condition.type
-        cond_args = condition.args
+        cond_args = copy.copy(condition.kwargs)
 
         try:
             typ = getattr(graph_scheduler.condition, cond_type)
@@ -1132,14 +1275,14 @@ class EvaluableGraph:
                     # value may be a string representing a TimeScale
                     cond_args[k] = getattr(
                         graph_scheduler.TimeScale,
-                        re.match(r"TimeScale\.(.*)", v).groups()[0],
+                        re.match(time_scale_str_regex, v).groups()[1],
                     )
                 except (AttributeError, IndexError, TypeError):
                     pass
             else:
                 cond_args[k] = new_v
 
-        substitute_condition_arguments(condition, typ)
+        cond_args = update_condition_arguments(cond_args, typ)
 
         try:
             return typ(**cond_args)
@@ -1173,9 +1316,6 @@ class EvaluableGraph:
                     )
 
 
-from modelspec.utils import FORMAT_NUMPY
-
-
 def main(example_file: str, array_format: str = FORMAT_NUMPY, verbose: bool = False):
     """
     Main entry point for execution engine.
