@@ -6,6 +6,7 @@ the future, the MDF should probably just compile to ONNX (or some other IR) for
 """
 import functools
 
+import torch
 import numpy as np
 import onnxruntime as ort
 import onnx.defs
@@ -72,6 +73,12 @@ def convert_type(v):
     if hasattr(v, "dtype") and v.dtype == np.int32:
         v = v.astype(np.int64)
 
+    if hasattr(v, "dtype") and v.dtype == np.float64:
+        v = v.astype(np.float32)
+
+    if isinstance(v, torch.Tensor):
+        v = v.detach().cpu().numpy()
+
     return v
 
 
@@ -116,12 +123,25 @@ def run_onnx_op(
             if cval.dtype != data.dtype:
                 inputs["constant_value"] = cval.astype(data.dtype)
 
+    # SkLearn ONNX doesn't seem to support ConcatFromSequence, see
+    # https://github.com/onnx/sklearn-onnx/issues/710
+    # Let us just use this implemetation I found in ONNX backend test.
+    if op_name == "ConcatFromSequence":
+        from onnx.backend.test.case.model.sequence import ConcatFromSequenceImpl
+
+        return {
+            "concat_result": ConcatFromSequenceImpl(
+                inputs["input_sequence"], **attributes
+            )
+        }
+
     op_class = import_class(f"skl2onnx.algebra.onnx_ops.Onnx{op_name}")
     input_names = list(inputs.keys())
     input_vals = list(inputs.values())
     op = op_class(
         *input_names, output_names=output_names, op_version=opset_version, **attributes
     )
+
     model_def = op.to_onnx(inputs)
     return predict_with_onnxruntime(model_def, *input_vals)
 
