@@ -200,6 +200,9 @@ class PortMapper:
 
         new_name = str(id).replace(".", "_")
 
+        # Remove :: from ids, these cause issues with parsing in the execution engine
+        new_name = new_name.replace("::", "_")
+
         # If the first character is a digit, precede with an underscore so this can never be interpreted
         # as number down the line.
         if new_name[0].isdigit():
@@ -319,9 +322,9 @@ def torchnode_to_mdfnode(
             # Try to get the shape and type of the input port
             inp_type = node.inputsAt(inp_i).type()
             try:
-                shape = str(inp_type.sizes()) if inp_type.sizes() else "(?)"
+                shape = tuple(inp_type.sizes()) if inp_type.sizes() else None
             except RuntimeError:
-                shape = "(?)"
+                shape = None
 
             mdf_node.input_ports.append(
                 InputPort(id=ip_name, shape=shape, type=str(inp_type))
@@ -405,7 +408,6 @@ def translate_graph(
 def pytorch_to_mdf(
     model: Union[Callable, torch.nn.Module, torch.ScriptFunction, torch.ScriptModule],
     args: Union[None, torch.Tensor, Tuple[torch.Tensor]] = None,
-    example_outputs: Union[None, torch.Tensor, Tuple[torch.Tensor]] = None,
     trace: bool = False,
     use_onnx_ops: bool = True,
 ) -> Union[Model, Graph]:
@@ -419,7 +421,6 @@ def pytorch_to_mdf(
         model: The model to translate into MDF.
         args: The input arguments for this model. If a nn.Module is passed then the model will be traced with these
             inputs. If a ScriptModule is passed, they are still needed to deterimine input shapes.
-        example_outputs: Example outputs from the model for determing output shapes.
         trace: Force the use of tracing to compile the model. The default is to use torch.jit.script
         use_onnx_ops: Use ONNX ops when possible, fallback to ATEN ops when not available. Default is True. If False,
             use only ATEN ops.
@@ -463,10 +464,8 @@ def pytorch_to_mdf(
     graph, params_dict, torch_out = _model_to_graph(
         model=jit_model if graph else model,
         args=args,
-        example_outputs=example_outputs,
-        do_constant_folding=False,
+        do_constant_folding=True,
         training=TrainingMode.EVAL,
-        _retain_param_name=True,
         operator_export_type=operator_export_type,
         dynamic_axes={},
     )
@@ -510,5 +509,4 @@ if __name__ == "__main__":
     mdf_model, param_dict = pytorch_to_mdf(
         simple,
         args=(torch.tensor(1.0), torch.tensor(2.0)),
-        example_outputs=torch.tensor(0.0),
     )
